from abc import abstractmethod
from datetime import datetime
from typing import Set

import dash_core_components as dcc
import dash_daq as daq
import dash_html_components as html
from dash.dependencies import Output, Input, MATCH, ALL

from app import app
from components import CarSize
from interface import Interface
from pages.cost_trend.control import Control, Configuration, visible_style, invisible_style


class CarProperties(Control):
    def __init__(self, *args, **kwargs):
        super(CarProperties, self).__init__(*args, **kwargs)
        self._add_ctrl_element(title='Production Liability:',
                               ctrl_element=self._create_production_liability_slider('production_liability'))

        self._add_ctrl_element(title='Size:',
                               ctrl_element=self._create_size_slider('size'))
        self.add_parts()

    def _create_production_liability_slider(self, element_id):
        marks = {
            1: 'new',
            0: 'rescued'
        }
        return dcc.Slider(
            id=self._element_id(element_id),
            min=0, max=1, step=.1,
            value=self.config[element_id],
            marks=marks)

    def _create_size_slider(self, element_id):
        marks = {
            CarSize.LARGE: 'Large',
            CarSize.MEDIUM: 'Medium',
            CarSize.SMALL: 'Small'
        }
        return dcc.Slider(
            id=self._element_id(element_id),
            min=50, max=180, step=10,
            value=self.config[element_id],
            marks=marks)

    @abstractmethod
    def add_parts(self):
        raise NotImplementedError

    def _add_charging_interface(self, element_list=None):
        self._add_ctrl_element(title='Charging Efficiency [%]:',
                               ctrl_element=self._create_slider(
                                   element_id='charging_efficiency',
                                   lower=.5, upper=1, mark_every=.1, step=.02,
                                   mark_processing=lambda v: f'{int(v * 100)}%'),
                               element_list=element_list
                               )

    def _add_motor_interface(self, element_list=None, skip: Set[str] = None):
        if skip is None:
            skip = set()

        if not 'efficiency_motor' in skip:
            self._add_ctrl_element(title='Efficiency [kWh/100km]:',
                                   ctrl_element=self._create_slider(
                                       element_id='efficiency_motor',
                                       lower=10, upper=30, mark_every=2, step=.5),
                                   element_list=element_list)

        if not 'min_battery_health' in skip:
            self._add_ctrl_element(title='Minimal Battery Health [%]:',
                                   ctrl_element=self._create_slider(
                                       element_id='min_battery_health',
                                       lower=0, upper=1, mark_every=.1, step=.01,
                                       mark_processing=lambda v: f'{int(100 * v)}%'),
                                   element_list=element_list
                                   )

        if not 'millage_motor' in skip:
            self._add_ctrl_element(title='Initial Mileage (E-Motor) [km]:',
                                   ctrl_element=self._create_slider(
                                       element_id='millage_motor',
                                       lower=0, upper=1000000, mark_every=100000, step=10000,
                                       mark_processing=lambda v: f'{int(v / 1000)}k'),
                                   element_list=element_list
                                   )

        if not 'lifetime_motor' in skip:
            self._add_ctrl_element(title='Lifetime (E-Motor) [km]:',
                                   ctrl_element=self._create_slider(
                                       element_id='lifetime_motor',
                                       lower=0, upper=1000000, mark_every=100000, step=10000,
                                       mark_processing=lambda v: f'{int(v / 1000)}k'),
                                   element_list=element_list
                                   )

    def _add_battery_interface(self, element_list=None):
        self._add_ctrl_element(title='Capacity [kWh]:',
                               ctrl_element=self._create_slider(
                                   element_id='capacity',
                                   lower=0, upper=150, mark_every=10, step=1),
                               element_list=element_list
                               )

        today = datetime.today()
        now = today.year + (today.month - 1) * 1 / 12

        self._add_ctrl_element(title='Manufacturing Time of Battery:',
                               ctrl_element=self._create_slider(
                                   element_id='battery_manufacturing',
                                   lower=2010, upper=now, mark_every=1, step=1 / 12),
                               element_list=element_list
                               )
        self._add_ctrl_element(title='Battery Technology:',
                               ctrl_element=self._create_dropdown(
                                   element_id='battery_technology',
                                   options=[
                                       {'label': technology, 'value': technology}
                                       for technology in Interface.available_batteries()]),
                               element_list=element_list
                               )

    def _add_transformation_interface(self):
        self._add_ctrl_element(
            title='Transformation to BEV',
            ctrl_element=daq.BooleanSwitch(
                id=self._element_id('transformation_switch'),
                on=self.config['transformation']
            )
        )

        transform_elements = []
        self._add_ctrl_element(
            title='in km',
            ctrl_element=dcc.Input(
                value=self.config['transformation_km'],
                inputMode='numeric',
                id=self._element_id('transformation_km')),
            element_list=transform_elements
        )
        self._add_motor_interface(element_list=transform_elements)
        self._add_battery_interface(element_list=transform_elements)
        self._add_charging_interface(element_list=transform_elements)

        self._elements.append(html.Div(
            transform_elements,
            id=self._element_id(
                property_name='transformation_properties',
                role='property_container'
            )
        ))

    def _add_ice_interface(self):
        self._add_ctrl_element(title='Efficiency [l/100km]:',
                               container_id='efficiency_ice_l_100km',
                               ctrl_element=self._create_slider(
                                   element_id='efficiency_ice_l_100km',
                                   lower=0, upper=20, mark_every=1, step=.1)
                               )

        self._add_ctrl_element(title='Efficiency [kg/100km]:',
                               container_id='efficiency_ice_kg_100km',
                               ctrl_element=self._create_slider(
                                   element_id='efficiency_ice_kg_100km',
                                   lower=0, upper=20, mark_every=1, step=.1)
                               )

        self._add_ctrl_element(title='Initial Mileage (ICE) [km]:',
                               ctrl_element=self._create_slider(
                                   element_id='millage_ice',
                                   lower=0, upper=1000000, mark_every=100000, step=10000,
                                   mark_processing=lambda v: f'{int(v / 1000)}k')
                               )

        self._add_ctrl_element(title='Lifetime (ICE) [km]:',
                               ctrl_element=self._create_slider(
                                   element_id='lifetime_ice',
                                   lower=0, upper=1000000, mark_every=100000, step=10000,
                                   mark_processing=lambda v: f'{int(v / 1000)}k')
                               )

        self._add_ctrl_element(
            title='Fuel',
            ctrl_element=self._create_dropdown(
                element_id='fuel',
                options=[
                    {'label': fuel, 'value': fuel}
                    for fuel in Interface.available_ice_fuels()]
            )
        )

    def _add_fc_interface(self):
        self._add_ctrl_element(title='Efficiency [kg/100km]:',
                               ctrl_element=self._create_slider(
                                   element_id='efficiency_fc',
                                   lower=0, upper=3, mark_every=.2, step=.05,
                               )
                               )

        self._add_ctrl_element(title='Power [kW]:',
                               ctrl_element=self._create_slider(
                                   element_id='power_fc',
                                   lower=0, upper=100, mark_every=10, step=1,
                               )
                               )

    def _add_hydrogen_tank_interface(self):
        self._add_ctrl_element(title='Hydrogen Capacity [kg]:',
                               ctrl_element=self._create_slider(
                                   element_id='hydrogen_capacity_kg',
                                   lower=0, upper=10, mark_every=1, step=.1,
                                   mark_processing=lambda v: f'{v}kg')
                               )


class ICEProperties(CarProperties):
    def __init__(self, *args, **kwargs):
        super(ICEProperties, self).__init__(*args, **kwargs)

    def add_parts(self):
        self._add_ice_interface()
        self._add_transformation_interface()


class BEVProperties(CarProperties):
    def __init__(self, *args, **kwargs):
        super(BEVProperties, self).__init__(*args, **kwargs)

    def add_parts(self):
        self._add_battery_interface()
        self._add_charging_interface()
        self._add_motor_interface()


class PHEVProperties(CarProperties):
    def __init__(self, *args, **kwargs):
        super(PHEVProperties, self).__init__(*args, **kwargs)

    def add_parts(self):
        self._add_ice_interface()
        self._add_battery_interface()
        self._add_charging_interface()
        self._add_motor_interface()


class FCEVProperties(CarProperties):
    def __init__(self, *args, **kwargs):
        super(FCEVProperties, self).__init__(*args, **kwargs)

    def add_parts(self):
        self._add_fc_interface()
        self._add_hydrogen_tank_interface()
        self._add_motor_interface(skip={'efficiency_motor'})
        self._add_battery_interface()


class CarConfiguration(Configuration):
    def default_values(self):
        today = datetime.today()
        now = today.year + (today.month - 1) * 1 / 12
        return {
            'efficiency_motor': 18,
            'min_battery_health': .8,
            'charging_efficiency': .9,
            'millage_motor': 0,
            'lifetime_motor': 250000,
            'capacity': 50,
            'battery_manufacturing': now,
            'battery_technology': Interface.available_batteries()[0],
            'efficiency_ice_kg_100km': 7,
            'efficiency_ice_l_100km': 7,
            'lifetime_ice': 250000,
            'millage_ice': 0,
            'efficiency_fc': .8,
            'size': 100,
            'production_liability': 1,
            'hydrogen_capacity_kg': 2.5,  # https://de.wikipedia.org/w/index.php?title=Toyota_Mirai&oldid=201939586
            'power_fc': 40,
            'transformation_km': 1000,
            'transformation': False,
            'fuel': Interface.available_ice_fuels()[0]
        }

    def key(self):
        return 'car'

    def factories(self):
        return {
            'BEV': BEVProperties,
            'ICE': ICEProperties,
            'PHEV': PHEVProperties,
            'FCEV': FCEVProperties
        }

    def create_table_row(self, entry_id, type):
        name = f'new {type} vehicle'
        return {'type': type, 'name': name, 'id': entry_id}

    def title(self):
        return 'Cars'

    def table_cols(self):
        return [
            {'id': 'name', 'name': 'Name', 'editable': True},
            {'id': 'type', 'name': 'Type', 'editable': False}
        ]

    def additional_config_inputs(self):
        return [Input({
            'role': 'setup_element',
            'id': ALL,
            'property': 'transformation_switch',
            'key': self.key()
        }, 'on')]

    def consume_extra_data(self, inputs, contexts, ids, config):
        transform_button_inputs = inputs[0]
        transform_button_contexts = contexts[0]
        for on, context in zip(transform_button_inputs, transform_button_contexts):
            car_id = context['id']['id']
            if car_id not in ids:
                continue
            idx = ids.index(car_id)
            config[idx]['transformation'] = on
        return config

    def init_callbacks(self):
        super(CarConfiguration, self).init_callbacks()

        @app.callback(
            inputs=[Input({
                'role': 'setup_element',
                'id': MATCH,
                'property': 'transformation_switch',
                'key': self.key()
            }, 'on')],
            output=Output({
                'role': 'property_container',
                'id': MATCH,
                'property': 'transformation_properties',
                'key': self.key()
            }, 'style')
        )
        def enable_transform(on):
            return visible_style if on else invisible_style

        @app.callback(
            inputs=[Input({
                'role': 'setup_element',
                'id': MATCH,
                'property': 'fuel',
                'key': self.key()
            }, 'value')],
            output=[
                Output({
                    'role': 'setup_container',
                    'id': MATCH,
                    'property': 'efficiency_ice_l_100km',
                    'key': self.key()
                }, 'style'),
                Output({
                    'role': 'setup_container',
                    'id': MATCH,
                    'property': 'efficiency_ice_kg_100km',
                    'key': self.key()
                }, 'style')
            ]
        )
        def switch_efficiency_slider(fuel):
            liquid = Interface.is_liquid_fuel(fuel)
            kg_100km_visibility = invisible_style if liquid else visible_style
            l_100km_visibility = visible_style if liquid else invisible_style
            return l_100km_visibility, kg_100km_visibility
