import json
import uuid
from abc import ABC, abstractmethod
from json.decoder import JSONDecodeError

import dash
import dash_core_components as dcc
import dash_html_components as html
import dash_table
from dash.dependencies import Output, Input, State, MATCH, ALL

from app import app


class Control:
    def __init__(self, id, key, config):
        self.config = config
        self.id = id
        self._elements = list()
        self.key = key

    def elements(self):
        return self._elements

    def _element_id(self, property_name, role='setup_element'):
        return {
            'id': self.id,
            'key': self.key,
            'role': role,
            'property': property_name
        }

    def _add_ctrl_element(self, title, ctrl_element, element_list=None, container_id=None):
        elements = list()

        elements.append(html.Div(title))
        elements.append(html.Div(ctrl_element))
        element_list = element_list if element_list is not None else self._elements
        if container_id:
            container_id = self._element_id(container_id, role='setup_container')
            element_list.append(html.Div(elements, id=container_id))
        else:
            element_list.append(html.Div(elements))

    def _create_slider(self, element_id, lower, upper, mark_every, step, mark_processing=lambda v: str(v),
                       first_mark=None):
        marks = dict()
        first_mark = first_mark if first_mark is not None else lower
        for i in range(int((upper - first_mark) / mark_every) + 1):
            value = round(first_mark + mark_every * i, 4)
            marks[value] = mark_processing(value)

        return dcc.Slider(id=self._element_id(element_id),
                          min=lower, max=upper, step=step,
                          value=self.config[element_id], marks=marks)

    def _create_dropdown(self, element_id, options):
        return dcc.Dropdown(id=self._element_id(element_id),
                            options=options,
                            value=self.config[element_id],
                            clearable=False)


invisible_style = {'display': 'none'}
visible_style = {'display': 'block'}


class Configuration(ABC):
    @abstractmethod
    def title(self):
        raise NotImplementedError

    @abstractmethod
    def key(self):
        raise NotImplementedError

    @abstractmethod
    def factories(self):
        raise NotImplementedError

    def types(self):
        return self.factories().keys()

    @abstractmethod
    def create_table_row(self, entry_id, type):
        raise NotImplementedError

    @abstractmethod
    def table_cols(self):
        raise NotImplementedError

    @abstractmethod
    def default_values(self):
        raise NotImplementedError

    def create_properties(self, config):
        ctrl = self.factories()[config['type']](config['id'], key=self.key(), config=config)
        elements = ctrl.elements()
        element_id = {
            'role': 'property_setting',
            'id': config['id'],
            'key': self.key()}
        controls = html.Div(elements, id=element_id, style={'display': 'none'})
        # config_entries = ctrl.config
        return controls

    def setup(self):
        elements = list()
        headers = []
        headers.append(html.Div(
            html.H4(self.title()),
            id=f'{self.key()}_config_header',
            className='config_header')
        )

        buttons = list()

        for t in self.types():
            id = {
                'key': self.key(),
                'role': 'add_button',
                'type': t
            }
            text = '+'
            if not t == '':
                text += f' {t}'

            button = html.Button(text,
                                 id=id,
                                 n_clicks_timestamp=0,
                                 className='add_button')
            buttons.append(button)

        headers.append(html.Div(buttons,
                                id=f'{self.key()}_buttons',
                                className='add_button_container'))

        elements.append(html.Div(headers,
                                 id=f'{self.key()}_top_row',
                                 className='top_row'))

        table = dash_table.DataTable(
            id=f'{self.key()}_table',
            row_selectable='single',
            row_deletable=True,
            columns=self.table_cols(),
            data=[]
        )

        elements.append(table)
        elements.append(html.Div(id=f'{self.key()}_config',
                                 style={'display': 'none'}))
        elements.append(html.Div(id=f'{self.key()}_property_container',
                                 className='property_container'))

        return html.Div(elements,
                        id=f'{self.key()}_setup_container',
                        className='setup_container')

    def additional_config_inputs(self):
        return []

    def consume_extra_data(self, inputs, context, ids, config):
        return config

    def init_callbacks(self):
        # ADD BUTTON | LOAD ==> TABLE
        @app.callback(
            output=Output(f'{self.key()}_table', 'data'),
            state=[State(f'{self.key()}_table', 'data')],
            inputs=[Input({'key': self.key(),
                           'role': 'add_button',
                           'type': ALL
                           }, 'n_clicks_timestamp'),
                    Input('loaded_data', 'children')
                    ]
        )
        def car_table(time_stamps, loaded, data):
            for input in dash.callback_context.inputs_list[0]:
                if trigger_id() == input['id']:
                    entry_id = str(uuid.uuid1())  # RANDOM TODO better by time, so really NO duplicaes
                    element_type = input['id']['type']
                    data += [self.create_table_row(entry_id, type=element_type)]

            ids = [row['id'] for row in data]

            if trigger_id() == 'loaded_data' and loaded:
                loaded = json.loads(loaded)
                for entry in loaded[self.key()]:
                    entry_id = entry['id']
                    if entry_id in ids:
                        idx = ids.index(entry_id)
                        data[idx] = entry
                    else:
                        data.append(entry)

            return data

        # VISIBILITY OF PROPERTIES SECTIONS
        @app.callback(
            output=Output({
                'role': 'property_setting',
                'id': MATCH,
                'key': self.key()
            }, 'style'),
            state=[State(f'{self.key()}_table', 'data')],
            inputs=[Input(f'{self.key()}_table', 'selected_rows')]
        )
        def visible(selected_rows, table):
            if not selected_rows:
                return invisible_style

            selected_row = selected_rows[0]
            car_id = dash.callback_context.outputs_list['id']['id']
            selected_id = table[selected_row]['id']
            return visible_style if selected_id == car_id else invisible_style

        # CONFIG
        @app.callback(
            output=[
                Output(f'{self.key()}_config', 'children'),
                Output(f'{self.key()}_property_container', 'children')
            ],
            state=[
                State(f'{self.key()}_config', 'children'),
                State(f'{self.key()}_property_container', 'children'),
                State('loaded_data', 'children')
            ],
            inputs=[
                       Input(f'{self.key()}_table', 'data'),
                       Input({
                           'role': 'setup_element',
                           'id': ALL,
                           'property': ALL,
                           'key': self.key()
                       }, 'value')
                   ] + self.additional_config_inputs()
        )
        def car_config(*inputs):
            table_data = inputs[0]
            properties = inputs[1]
            extra_inputs = inputs[2:-3]
            extra_inputs_context = dash.callback_context.inputs_list[2:]
            state = inputs[-3]
            property_children = inputs[-2]
            loaded = inputs[-1]

            loaded = json.loads(loaded)[self.key()] if loaded else {}
            loaded_ids = [row['id'] for row in loaded]

            if not property_children:
                property_children = []

            if not state:
                state = {}
            else:
                state = json.loads(state)
            ids = state.get('ids', [])
            entries = state.get('entries', [])

            table_content = {
                row['id']: row for row in table_data
            }

            deleted = set(ids) - set(table_content.keys())
            for entry in deleted:
                idx = ids.index(entry)
                del entries[idx]
                del property_children[idx]
                del ids[idx]

            for entry in table_content:
                if entry not in ids:
                    entry_type = table_content[entry]['type']

                    ids.append(entry)
                    config = self.default_values()
                    if entry in loaded_ids:
                        config.update(loaded[loaded_ids.index(entry)])
                    else:
                        config.update({
                            'id': entry,
                            'type': entry_type
                        })

                    controls = self.create_properties(config)

                    property_children.append(controls)
                    entries.append(config)

                idx = ids.index(entry)
                entries[idx].update(table_content[entry])

            entries = self.consume_extra_data(extra_inputs, extra_inputs_context, ids, entries)

            for value, context in zip(properties, dash.callback_context.inputs_list[1]):
                entry_id = context['id']['id']
                if not entry_id in ids:
                    continue
                property = context['id']['property']
                idx = ids.index(entry_id)
                entries[idx][property] = value

            return json.dumps({'ids': ids, 'entries': entries}), property_children


def trigger_id():
    trigger_id = dash.callback_context.triggered[0]['prop_id'].split('.')[0]
    if not trigger_id:
        return None
    try:
        trigger_id = json.loads(trigger_id)
    except JSONDecodeError:
        pass
    return trigger_id


def did_input_trigger(input_idx):
    input_id = dash.callback_context.inputs_list[input_idx]['id']
    return input_id == trigger_id()


# from https://dash.plotly.com/datatable/conditional-formatting
def data_bars(data, column):
    if len(data) == 0:
        return []

    values = [row[column] for row in data]

    n_bins = 100
    bounds = [i * (1.0 / n_bins) for i in range(n_bins + 1)]
    ranges = [
        max(values) * i
        for i in bounds
    ]
    styles = []
    for i in range(1, len(bounds)):
        min_bound = ranges[i - 1]
        max_bound = ranges[i]
        max_bound_percentage = bounds[i] * 100

        bar_color = '#0074D9'
        styles.append({
            'if': {
                'filter_query': (
                        f'{{{column}}} >= {min_bound}' +
                        (f' && {{{column}}} < {max_bound}' if (i < len(bounds) - 1) else '')
                ),
                'column_id': column
            },
            'background': (
                f'linear-gradient('
                f'90deg, '
                f'{bar_color} 0%, {bar_color} {round(max_bound_percentage)}%,'
                f' white {round(max_bound_percentage)}%, white 100%)'
            ),
            'paddingBottom': 2,
            'paddingTop': 2
        })
    return styles
