import dash_html_components as html
import dash_table
from dash.dependencies import Output, Input, State, MATCH, ALL
from dash_table import FormatTemplate
from dash_table.Format import Sign, Format, Scheme, Symbol

from app import app
from pages.cost_trend.control import Configuration, Control, data_bars, did_input_trigger


class DriverProperties(Control):
    def __init__(self, *args, **kwargs):
        super(DriverProperties, self).__init__(*args, **kwargs)
        profile = DrivingProfileSetup(self.id, config=self.config)
        self._add_ctrl_element(title='Driving Profile:', ctrl_element=profile.elements())
        self.config.update(profile.config)
        self.config['base_time'] = 2020


class DrivingProfileSetup:
    def __init__(self, driver_id, config):
        self.driver_id = driver_id
        self.config = config

    def elements(self):
        add_button = html.Button('+ Trip',
                                 id={'role': 'profile_table_add',
                                     'id': self.driver_id},
                                 n_clicks=0,
                                 className='add_button')

        profile = [{'trip': int(trip), 'count': int(count)} for trip, count in self.config['profile'].items()]

        table = dash_table.DataTable(
            id={
                'role': 'profile_table',
                'id': self.driver_id
            },
            row_deletable=True,
            # cell_selectable=False,
            columns=[
                {'id': 'trip',
                 'name': 'Trip Length',
                 'editable': True,
                 'type': 'numeric',
                 'format': Format(
                     precision=0,
                     scheme=Scheme.fixed,
                     symbol=Symbol.yes,
                     symbol_suffix='km')
                 },
                {'id': 'count',
                 'name': 'Trip Count',
                 'editable': True,
                 'type': 'numeric'
                 },
                {'id': 'count_share',
                 'name': 'Count Share',
                 'editable': False,
                 'type': 'numeric',
                 'format': FormatTemplate.percentage(1).sign(Sign.default)
                 },
                {'id': 'km',
                 'name': 'total km',
                 'editable': False,
                 'type': 'numeric',
                 'format': Format(
                     precision=0,
                     scheme=Scheme.fixed,
                     symbol=Symbol.yes,
                     symbol_suffix='km')
                 },
                {'id': 'km_share',
                 'name': 'km Share',
                 'editable': False,
                 'type': 'numeric',
                 'format': FormatTemplate.percentage(1).sign(Sign.default)
                 }
            ],
            data=profile,
            sort_action='native',
            sort_by=[{
                'column_id': 'trip',
                'direction': 'asc'
            }]
        )

        km_sum = html.Div(id={'role': 'profile_km_sum', 'id': self.driver_id})

        return [add_button, table, km_sum]


class DriverConfiguration(Configuration):
    def default_values(self):
        return {  # TODO put better default scenario
            'profile': {
                '10': 10,
                '20': 5
            }
        }

    def table_cols(self):
        return [
            {'id': 'name', 'name': 'Name', 'editable': True},
            {'id': 'yearly_km', 'name': 'Yearly Km', 'editable': True, 'type': 'numeric'},
        ]

    def key(self):
        return 'driver'

    def factories(self):
        return {
            '': DriverProperties
        }

    def create_table_row(self, entry_id, type):
        name = f'new driver'
        return {'yearly_km': '20000',
                'name': name,
                'id': entry_id,
                'type': ''}

    def title(self):
        return 'Drivers'

    def additional_config_inputs(self):
        return [Input({'role': 'profile_table',
                       'id': ALL},
                      'data')]

    def consume_extra_data(self, inputs, context, ids, config):
        for driver_table_context, profile in zip(context[0], inputs[0]):
            driver_id = driver_table_context['id']['id']
            if driver_id not in ids:
                continue
            idx = ids.index(driver_id)

            reduced_profile = {row['trip']: row['count'] for row in profile}
            config[idx]['profile'] = reduced_profile
        return config

    def init_callbacks(self):
        super(DriverConfiguration, self).init_callbacks()

        @app.callback(
            output=Output({'role': 'profile_km_sum',
                           'id': MATCH},
                          'children'),
            inputs=[Input({'role': 'profile_table',
                           'id': MATCH},
                          'data')]
        )
        def add_table(data):
            summed = 0
            for row in data:
                summed += float(row['trip']) * float(row['count'])
            return f'Summed: {summed} km'

        # ADD TRIP TO PROFILE
        @app.callback(
            output=[Output({'role': 'profile_table',
                            'id': MATCH},
                           'data'),
                    Output({'role': 'profile_table',
                            'id': MATCH},
                           'style_data_conditional')
                    ],
            state=[State({'role': 'profile_table',
                          'id': MATCH},
                         'data'),
                   ],
            inputs=[Input({'role': 'profile_table_add',
                           'id': MATCH},
                          'n_clicks'),
                    Input({'role': 'profile_table',
                           'id': MATCH},
                          'data_timestamp')
                    ]
        )
        def add_table(n_clicks, table_timestamp, data):
            if did_input_trigger(0):
                new_entry = {
                    'trip': 1,
                    'count': 1
                }
                data.append(new_entry)
            new_table = self.fill_profile_table(data)
            return new_table, data_bars(new_table, 'km_share') + data_bars(new_table, 'count_share')

    def fill_profile_table(self, data):
        sum_count = sum([d['count'] for d in data])
        sum_kms = sum([d['count'] * d['trip'] for d in data])

        return [
            {
                'count': d['count'],
                'trip': d['trip'],
                'count_share': d['count'] / sum_count,
                'km': d['trip'] * d['count'],
                'km_share': d['trip'] * d['count'] / sum_kms
            }
            for d in data
        ]
