import base64
import json
import urllib.parse

import dash_core_components as dcc
import dash_html_components as html
import plotly.graph_objects as go
from dash.dependencies import Output, Input, State

from app import app
from interface import Interface, NoProperScenario
from pages.cost_trend.car_control import CarConfiguration
from pages.cost_trend.driver_control import DriverConfiguration
from pages.cost_trend.scenario_control import ScenarioConfiguration


class TrendLayout:
    def __init__(self):
        self.car_config = CarConfiguration()
        self.driver_config = DriverConfiguration()
        self.scenario_config = ScenarioConfiguration()

    def request(self, cars, drivers, scenarios):
        return {
            'car': json.loads(cars)['entries'],
            'driver': json.loads(drivers)['entries'],
            'scenario': json.loads(scenarios)['entries']
        }

    def create_data_uri(self, content):
        return 'data:text/txt;charset=utf-8,' + urllib.parse.quote(content, safe='')

    def init_callback(self):
        to_km = 250000
        self.car_config.init_callbacks()
        self.driver_config.init_callbacks()
        self.scenario_config.init_callbacks()

        def update_trend_graph(emissions):
            if not emissions:
                data = []
            else:
                data = [
                    {
                        'x': list(graph['accumulated'].keys()),
                        'y': list(graph['accumulated'].values()),
                        'name': graph['name'],
                        'type': 'scatter'
                    } for scenario_id, graph in emissions.items()
                ]

            return {
                'data': data,
                'layout': {
                    'uirevision': 'hallo daniel'
                }
            }

        @app.callback(output=Output('loaded_data', 'children'),
                      state=[State('last_data', 'data')],
                      inputs=[Input('load_data', 'contents')])
        def update_output(contents, last_data):
            if contents:
                content_type, content_string = contents.split(',')
                decoded = base64.b64decode(content_string)
                return decoded.decode('utf-8')
            else:
                return last_data

        @app.callback(
            output=Output('last_data', 'data'),
            inputs=[Input('car_config', 'children'),
                    Input('driver_config', 'children'),
                    Input('scenario_config', 'children')])
        def save_state(cars, drivers, scenarios):
            data = self.request(cars, drivers, scenarios)
            return json.dumps(data)

        @app.callback(
            output=Output('download-link', 'href'),
            inputs=[Input('car_config', 'children'),
                    Input('driver_config', 'children'),
                    Input('scenario_config', 'children')])
        def update_download_link(cars, drivers, scenarios):
            data = self.request(cars, drivers, scenarios)
            return self.create_data_uri(json.dumps(data))

        @app.callback(
            output=Output('export_json_link', 'href'),
            state=[
                State('car_config', 'children'),
                State('driver_config', 'children'),
                State('scenario_config', 'children')
            ],
            inputs=[Input('update_button', 'n_clicks')]
        )
        def update_download_link(update_clicks, cars, drivers, scenarios):
            data = ''
            if update_clicks:
                interface = Interface(self.request(cars, drivers, scenarios))
                data = json.dumps(interface.details(to_km))

            return self.create_data_uri(data)

        @app.callback(
            output=Output('effects_json_link', 'href'),
            inputs=[Input('calculations', 'children')]
        )
        def update_download_link(caluclations):
            return self.create_data_uri(caluclations)

        @app.callback(
            output=Output('export_csv_link', 'href'),
            state=[
                State('car_config', 'children'),
                State('driver_config', 'children'),
                State('scenario_config', 'children')
            ],
            inputs=[Input('update_button', 'n_clicks')]
        )
        def update_download_link(update_clicks, cars, drivers, scenarios):
            data = ''
            if update_clicks:
                interface = Interface(self.request(cars, drivers, scenarios))
                details = interface.details(to_km)

                xs = None
                lines = []

                for name, graph in details['accumulated'].items():
                    lines.append([f'{name} (accumulated)'] + [str(y) for y in graph.values()])
                    xs = [str(x) for x in graph.keys()]

                for name, graphs in details['detailed'].items():
                    for efx_name, data in graphs.items():
                        lines.append([f'{name} ({efx_name})'] + [str(y) for y in data.values()])

                lines = [['x'] + xs] + lines
                data = '\n'.join([','.join(l) for l in lines])
            return self.create_data_uri(data)

        @app.callback(
            output=Output('cost_flow', 'figure'),
            state=[
                State('car_config', 'children'),
                State('driver_config', 'children'),
                State('scenario_config', 'children'),
                State('calculations', 'children')
            ],
            inputs=[
                Input('costs_scatter', 'hoverData'),
                Input('scenario_table', 'selected_rows')
            ])
        def update_sankey(hover_data, selected_scenario, cars, drivers, scenarios, calculations):
            if cars and drivers and scenarios and selected_scenario and calculations and hover_data:
                interface = Interface(self.request(cars, drivers, scenarios))
                scenarios = json.loads(scenarios)['entries']
                scenario_idx = scenarios[selected_scenario[0]]['id']
                effects = json.loads(calculations).get('effects')

                x = hover_data['points'][0]['x']
                try:
                    data = interface.sankey(x=x, scenario_id=scenario_idx, effects=effects)
                except (NoProperScenario):
                    data = {}
            else:
                data = {}

            return go.Figure(data=[go.Sankey(
                node=dict(
                    pad=15,
                    thickness=20,
                    line=dict(color="black", width=0.5),
                    label=data.get('labels'),
                    color="blue"
                ),
                link=dict(
                    source=data.get('source'),  # indices correspond to labels, eg A1, A2, A2, B1, ...
                    target=data.get('target'),
                    value=data.get('value')
                ))])

        @app.callback(
            output=[
                Output('calculations', 'children'),
                Output('costs_scatter', 'figure')
            ],
            state=[
                State('calculations', 'children')
            ],
            inputs=[
                Input('car_config', 'children'),
                Input('driver_config', 'children'),
                Input('scenario_config', 'children')
            ]
        )
        def calculate(cars, drivers, scenarios, data):
            data = json.loads(data) if data else {}

            if cars and drivers and scenarios:
                interface = Interface(self.request(cars, drivers, scenarios))
                old_effects = data.get('effects', dict())
                data['effects'] = interface.effects(old_effects=old_effects, km=to_km)
                old_emissions = data.get('emissions', dict())
                data['emissions'] = interface.emissions(effects=data['effects'], old_emissions=old_emissions)

            return json.dumps(data), update_trend_graph(data.get('emissions'))

    def scatter_layout(self):
        costs_scatter = dcc.Graph(
            figure={}, id='costs_scatter'
        )
        return html.Div([costs_scatter], id='costs_scatter_container')

    def cost_flow_layout(self):
        flow_diagram = dcc.Graph(
            figure={},
            id='cost_flow')
        return html.Div([flow_diagram], id='cost_flow_container')

    def layout(self):
        diagrams = html.Div([self.scatter_layout(), self.cost_flow_layout()], id='diagrams')

        setup = html.Div([
            self.car_config.setup(),
            self.scenario_config.setup(),
            self.driver_config.setup()
        ], id='setup')
        container = html.Div([diagrams, setup], id='container')

        content = list()
        content.append(html.H1('Vehicle Estimator'))
        content.append(container)
        # content.append(plant_dropdown(element_id='plant_selection'))
        footer = list()
        footer.append(html.Div(None, id='loaded_data', style={'display': 'none'}))
        footer.append(html.Div(None, id='calculations', style={'display': 'none'}))
        footer.append(dcc.Upload(
            id='load_data',
            children=html.Div([
                html.A('Load Data')
            ])))
        footer.append(html.Div(html.A(
            'Save Data',
            id='download-link',
            download='vehicle_estimator.json',
            href='',
            target='_blank'
        )))
        footer.append(html.Div(html.A(
            'Effects (JSON)',
            id='effects_json_link',
            download='effects.json',
            href='',
            target='_blank'
        )))
        footer.append(dcc.Store(
            id='last_data',
            data=None,
            storage_type='local'
        ))
        footer.append(html.Button('Update Export', id='update_button'))
        footer.append(html.Div(html.A(
            'Export (JSON)',
            id='export_json_link',
            download='graphes.json',
            href='',
            target='_blank'
        )))
        footer.append(html.Div(html.A(
            'Export (CSV)',
            id='export_csv_link',
            download='graphes.csv',
            href='',
            target='_blank'
        )))
        footer.append(html.Div('Special thanks and credits to Daniel.'))
        content.append(html.Div(footer, id='footer_container'))
        return html.Div(content, id='all')
