import json
from abc import abstractmethod, ABC

import dash_html_components as html
import dash_table
from dash.dependencies import Output, Input, MATCH, ALL, State
from dash_table import FormatTemplate
from dash_table.Format import Sign

from app import app
from interface import Interface
from pages.cost_trend.control import Configuration, Control, did_input_trigger, data_bars


class ScenarioCarSetup(ABC):
    def __init__(self, scenario_id, config):
        self.scenario_id = scenario_id
        self.config = config

    @abstractmethod
    def type(self):
        raise NotImplementedError

    def elements(self):
        add_button = html.Button('+ Vehicle',
                                 id={'role': 'scenario_car_add',
                                     'id': self.scenario_id,
                                     'type': self.type()},
                                 n_clicks=0,
                                 className='add_button')

        table = self.create_table()
        return [add_button, table]

    @abstractmethod
    def create_table(self):
        raise NotImplementedError


class FixScenarioCarSetup(ScenarioCarSetup):
    def __init__(self, *args, **kwargs):
        super(FixScenarioCarSetup, self).__init__(*args, **kwargs)

    def type(self):
        return 'fixed'

    def create_table(self):
        return dash_table.DataTable(
            id={
                'role': 'scenario_car_table',
                'type': self.type(),
                'id': self.scenario_id
            },
            row_deletable=True,
            # cell_selectable=False,
            columns=[
                {'id': 'car',
                 'name': 'Vehicle',
                 'editable': True,
                 'presentation': 'dropdown'
                 },
                {'id': 'share',
                 'name': 'Share',
                 'editable': True,
                 'type': 'numeric'
                 },
                {'id': 'percentage',
                 'name': 'Percentage',
                 'editable': False,
                 'type': 'numeric',
                 'format': FormatTemplate.percentage(1).sign(Sign.default)
                 }
            ],
            # editable=True,
            data_timestamp=0,
            data=self.config['cars']
        )


class DynamicScenarioCarSetup(ScenarioCarSetup):
    def __init__(self, *args, **kwargs):
        super(DynamicScenarioCarSetup, self).__init__(*args, **kwargs)

    def type(self):
        return 'dynamic'

    def create_table(self):
        return dash_table.DataTable(
            id={
                'role': 'scenario_car_table',
                'type': self.type(),
                'id': self.scenario_id
            },
            # editable=True,
            row_deletable=True,
            # cell_selectable=False,
            columns=[
                {'id': 'user_order',
                 'name': 'Order',
                 'editable': True,
                 'presentation': 'dropdown'
                 },
                {'id': 'car',
                 'name': 'Vehicle',
                 'editable': True,
                 'presentation': 'dropdown'
                 },
                {'id': 'charges',
                 'name': 'Charging Stops',
                 'editable': True,
                 'type': 'numeric'
                 }
            ],
            data_timestamp=0,
            data=self.config['cars'],
            sort_action='native',
            sort_by=[{
                'column_id': 'user_order',
                'direction': 'asc'
            }]
        )


class ScenarioControl(Control):
    def __init__(self, *args, **kwargs):
        super(ScenarioControl, self).__init__(*args, **kwargs)
        self._add_ctrl_element(
            title='Hydrogen Source',
            ctrl_element=self._create_dropdown(
                element_id='hydrogen_source',
                options=[
                    {'label': src, 'value': src}
                    for src in Interface.available_hydrogen_sources()
                ]
            )
        )
        self._add_ctrl_element(
            title='Diesel Source',
            ctrl_element=self._create_dropdown(
                element_id='diesel_source',
                options=[
                    {'label': src, 'value': src}
                    for src in Interface.available_diesel_sources()
                ]
            )
        )
        self._add_ctrl_element(
            title='Petrol Source',
            ctrl_element=self._create_dropdown(
                element_id='petrol_source',
                options=[
                    {'label': src, 'value': src}
                    for src in Interface.available_petrol_sources()
                ]
            )
        )


class FixShareScenarioProperties(ScenarioControl):
    def __init__(self, *args, **kwargs):
        super(FixShareScenarioProperties, self).__init__(*args, **kwargs)
        table = FixScenarioCarSetup(scenario_id=self.id, config=self.config)
        self._add_ctrl_element('Vehicles:', table.elements())
        self.config.update(table.config)


class ManualHybridScenarioProperties(ScenarioControl):
    def __init__(self, *args, **kwargs):
        super(ManualHybridScenarioProperties, self).__init__(*args, **kwargs)
        table = DynamicScenarioCarSetup(scenario_id=self.id, config=self.config)
        self.config.update(table.config)
        self._add_ctrl_element('Vehicles:', table.elements())


class ScenarioConfiguration(Configuration):
    def default_values(self):
        return {
            'cars': [],
            'hydrogen_source': Interface.available_hydrogen_sources()[0],
            'diesel_source': 'crude oil',
            'petrol_source': 'crude oil'
        }

    def __init__(self):
        super(ScenarioConfiguration, self).__init__()
        self.default_plant = Interface.available_plants()[-1]

    def table_cols(self):
        return [
            {'id': 'name', 'name': 'Name', 'editable': True},
            {'id': 'driver', 'name': 'Driver', 'editable': True, 'presentation': 'dropdown'},
            {'id': 'plant', 'name': 'Power Supply', 'editable': True, 'presentation': 'dropdown'},
            {'id': 'type', 'name': 'Type', 'editable': False},
        ]

    def key(self):
        return 'scenario'

    def factories(self):
        return {
            'fix': FixShareScenarioProperties,
            'dynamic': ManualHybridScenarioProperties
        }

    def create_table_row(self, entry_id, type):
        name = f'new {type} scenario'
        return {'type': type,
                'name': name,
                'id': entry_id,
                'driver': '',
                'plant': self.default_plant}

    def title(self):
        return 'Scenarios'

    def additional_config_inputs(self):
        return [Input({'role': 'scenario_car_table',
                       'type': 'fixed',
                       'id': ALL},
                      'data'),
                Input({'role': 'scenario_car_table',
                       'type': 'dynamic',
                       'id': ALL},
                      'data')
                ]

    def consume_extra_data(self, inputs, context, ids, config):

        def update_cfg(config, fixed_scenario_context, fixed_scenario_inputs, ids):
            for scenario_table_context, cars_table in zip(fixed_scenario_context, fixed_scenario_inputs):
                scenario_id = scenario_table_context['id']['id']
                if scenario_id not in ids:
                    continue
                idx = ids.index(scenario_id)
                config[idx]['cars'] = cars_table

        fixed_scenario_context = context[0]
        fixed_scenario_inputs = inputs[0]
        update_cfg(config, fixed_scenario_context, fixed_scenario_inputs, ids)

        dynamic_scenario_context = context[1]
        dynamic_scenario_inputs = inputs[1]
        update_cfg(config, dynamic_scenario_context, dynamic_scenario_inputs, ids)

        return config

    def init_callbacks(self):
        super(ScenarioConfiguration, self).init_callbacks()

        # FILL DRIVER DROPDOWNS
        @app.callback(
            output=Output(f'{self.key()}_table', 'dropdown'),
            inputs=[Input(f'driver_config', 'children')]

        )
        def fill_driver_dropdown(driver_cfg):
            drivers = [{'label': row['name'], 'value': row['id']}
                       for row in json.loads(driver_cfg)['entries']]
            return {
                'driver': {
                    'options': drivers,
                    'clearable': False
                },
                'plant': {
                    'options': [{'label': plant, 'value': plant} for plant in Interface.available_plants()],
                    'clearable': True}
            }

        # FILL CAR DROPDOWNS
        @app.callback(
            output=Output({
                'role': 'scenario_car_table',
                'id': MATCH,
                'type': MATCH,
            }, 'dropdown_conditional'),
            state=[
                State({
                    'role': 'scenario_car_table',
                    'id': MATCH,
                    'type': MATCH
                }, 'data')
            ],
            inputs=[Input(f'car_config', 'children'),
                    ]
        )
        def fill_car_dropdown(car_cfg, table_data):
            all_cars = {row['id']: row['name']
                        for row in json.loads(car_cfg)['entries']}
            cars_in_use = [row['car'] for row in table_data if not row['car'] == '']
            free_cars = list(set(all_cars) - set(cars_in_use))
            result = [
                         {
                             'if': {
                                 'column_id': 'car',
                                 'filter_query': f'{{car}} eq "{car_id}"'
                             },
                             'options': [
                                 {'label': all_cars[car], 'value': car}
                                 for car in free_cars + [car_id]
                             ],
                             'clearable': False
                         }
                         for car_id in all_cars
                     ] + [
                         {
                             'if': {
                                 'column_id': 'car',
                                 'filter_query': '{car} eq \'\''
                             },
                             'options': [{'label': all_cars[car], 'value': car} for car in free_cars],
                             'clearable': False
                         }
                     ] + [
                         {
                             'if': {
                                 'column_id': 'user_order'
                             },
                             'options': [{'label': f'{i + 1}', 'value': i} for i in range(len(table_data))],
                             'clearable': False
                         }
                     ]

            return result

        # ADD CAR TO FIXED SCENARIO
        @app.callback(
            output=[Output({'role': 'scenario_car_table',
                            'id': MATCH,
                            'type': 'fixed'}, 'data'),
                    Output({'role': 'scenario_car_table',
                            'id': MATCH,
                            'type': 'fixed'}, 'style_data_conditional')],
            state=[State({'role': 'scenario_car_table',
                          'id': MATCH,
                          'type': 'fixed'}, 'data')],
            inputs=[Input({'role': 'scenario_car_add',
                           'id': MATCH,
                           'type': 'fixed'}, 'n_clicks'),
                    Input({'role': 'scenario_car_table',
                           'id': MATCH,
                           'type': 'fixed'}, 'data_timestamp'),
                    Input('car_config', 'children')])
        def add_car_to_fixed_sceanrio(n_clicks, data_timestamp, car_cfg, data):
            if did_input_trigger(0):
                new_entry = {
                    'car': '',
                    'share': 1
                }
                data.append(new_entry)

            sum_of_share = sum([d['share'] for d in data])

            new_data = [
                {
                    'share': d['share'],
                    'car': d['car'],
                    'percentage': d['share'] / sum_of_share
                }
                for d in data
            ]

            available_cars = json.loads(car_cfg)['ids']
            for row in new_data:
                if row['car'] not in available_cars:
                    row['car'] = ''

            return new_data, data_bars(new_data, 'percentage')

        # ADD CAR TO DYNAMIC SCENARIO
        @app.callback(
            output=Output({'role': 'scenario_car_table',
                           'id': MATCH,
                           'type': 'dynamic'}, 'data'),
            state=[State({'role': 'scenario_car_table',
                          'id': MATCH,
                          'type': 'dynamic'}, 'data')],
            inputs=[Input({'role': 'scenario_car_add',
                           'id': MATCH,
                           'type': 'dynamic'}, 'n_clicks'),
                    Input({'role': 'scenario_car_table',
                           'id': MATCH,
                           'type': 'dynamic'}, 'data_timestamp'),
                    Input('car_config', 'children')])
        def add_car_to_dynamic_scenario(n_clicks, data_timestamp, car_cfg, data):
            if did_input_trigger(0):
                new_entry = {
                    'car': '',
                    'charges': 0,
                    'user_order': 0,
                    'order': len(data)
                }
                data.append(new_entry)

            old_idx = None
            new_idx = None
            direction = None
            modified_row = None

            for row in data:
                if row['order'] != row['user_order']:
                    old_idx = row['order']
                    new_idx = row['user_order']
                    direction = -1 if old_idx > new_idx else 1
                    modified_row = row

            if direction:
                for row in data:
                    if min(old_idx, new_idx) <= row['order'] and row['order'] <= max(old_idx, new_idx):
                        row['order'] -= direction
                        row['user_order'] -= direction

                modified_row['order'] = new_idx
                modified_row['user_order'] = new_idx

            possible_idxs = range(len(data))
            used_idxs = [row['order'] for row in data]
            gaps = set(possible_idxs) - set(used_idxs)
            for gap in gaps:
                for row in data:
                    if row['order'] > gap:
                        row['order'] -= 1
                        row['user_order'] -= 1

            available_cars = json.loads(car_cfg)['ids']
            for row in data:
                if row['car'] not in available_cars:
                    row['car'] = ''
            return data
